//====================================================================
//Project Lynxmotion Phoenix
//
// Servo Driver - This version is setup to use the SSC-32 to control
// the servos.
//====================================================================
#if ARDUINO>99
#include <Arduino.h> // Arduino 1.0
#else
#include <Wprogram.h> // Arduino 0022
#endif
#include "Hex_Globals.h"
#include "ServoDriver.h"

#ifdef USE_SSC32

// Add support for running on non-mega Arduino boards as well.
#ifdef __AVR__
#if not defined(UBRR1H)
    SoftwareSerial SSCSerial(cSSC_IN, cSSC_OUT);
#endif
#endif

//=============================================================================
// Global - Local to this file only...
//=============================================================================

// definition of some helper functions
extern void SSCWrite(byte *pb, int ich);
extern int SSCRead (byte* pb, int cb, word wTimeout, word wEOL);


//--------------------------------------------------------------------
//Init
//--------------------------------------------------------------------
void ServoDriver::Init(void) {
    SSCSerial.begin(cSSC_BAUD);

}

//--------------------------------------------------------------------
//[GP PLAYER]
//--------------------------------------------------------------------
boolean ServoDriver::FIsGPEnabled(void)
{
    char abVer[40];        // give a nice large buffer.
    byte cbRead;

#ifdef __AVR__
#if not defined(UBRR1H)
    SSCSerial.listen();
#endif    
#endif
    SSCSerial.print("ver\r");
    
    cbRead = SSCRead((byte*)abVer, sizeof(abVer), 10000, 13);
    
    DBGSerial.write("Check GP Enable: ");
    if (cbRead > 0) {
        byte iT;
        for (iT = 0; iT < cbRead; iT++)
            DBGSerial.print(abVer[iT], HEX);
        DBGSerial.write((byte*)abVer, cbRead);
    }
    DBGSerial.print("\n\r");
        
    if ((cbRead > 3) && (abVer[cbRead-3]=='G') && (abVer[cbRead-2]=='P') && (abVer[cbRead-1]==13))
        return true;

    MSound (SOUND_PIN, 2, 40, 2500, 40, 2500);
    
    return false;
}


//--------------------------------------------------------------------
//[GP PLAYER]
void ServoDriver::GPPlayer(void)
{
    byte abStat[4];
    byte cbRead;
    
    //Start sequence
    if (GPStart==1) {
        AllowControllerInterrupts(false);    // If on xbee on hserial tell hserial to not processess...

        SSCSerial << "PL0SQ" << _DEC(GPSeq) << "ONCE\r"; //Start sequence
                delay(20);
                SSCSerial.flush();        // get rid of anything that was previously queued up...
    
        //Wait for GPPlayer to complete sequence    
        do {
            SSCSerial.print("QPL0\r");
            cbRead = SSCRead((byte*)abStat, sizeof(abStat), 10000, (word)-1);  //    [GPStatSeq, GPStatFromStep, GPStatToStep, GPStatTime]
#ifdef DEBUG_GP
            if (g_fDebugOutput)
                DBGSerial << "GP: " << _DEC(cbRead) << " " << _DEC(abStat[0]) << " " << _DEC(abStat[1])
                        << " " << _DEC(abStat[2]) << " " << _DEC(abStat[3]) << "\n\r";
#endif                        
            delay(20);
        }
        while ((cbRead == sizeof(abStat)) && ((abStat[0]!=255) || (abStat[1]!=0) || (abStat[2]!=0) || (abStat[3]!=0)));

        AllowControllerInterrupts(true);    // Ok to process hserial again...

        GPStart=0;
    }  
}

//--------------------------------------------------------------------
//[SERVO DRIVER Start] Updates the positions of the servos - This outputs
//         as much of the command as we can without committing it.  This
//         allows us to once the previous update was completed to quickly 
//        get the next command to start
//--------------------------------------------------------------------
#define cPwmDiv       991  //old 1059;
#define cPFConst      592  //old 650 ; 900*(1000/cPwmDiv)+cPFConst must always be 1500
                           // A PWM/deg factor of 10,09 give cPwmDiv = 991 and cPFConst = 592
                           // For a modified 5645 (to 180 deg travel): cPwmDiv = 1500 and cPFConst = 900.

void ServoDriver::UpdateServoDriver()
{        
    byte    LegIndex;
    word    wCoxaSSCV;        // Coxa value in SSC units
    word    wFemurSSCV;        //
    word    wTibiaSSCV;        //
#ifdef c4DOF
    word    wTarsSSCV;        //
#endif

    //Update Right Legs
    AllowControllerInterrupts(false);    // If on xbee on hserial tell hserial to not processess...

#ifdef DEBUG_GAITS
    if (g_fDebugOutput && g_fHexOn && (TravelLengthX || TravelLengthZ))
        DBGSerial << _DEC(GaitType) << " " << _DEC(GaitStep) << " " << _DEC(BalanceMode) << ":"; 
#endif
  

    for (LegIndex = 0; LegIndex <= 5; LegIndex++) {
        if (LegIndex < 3) {
            wCoxaSSCV = ((long)(-CoxaAngle1[LegIndex] +900))*1000/cPwmDiv+cPFConst;
            wFemurSSCV = ((long)(-FemurAngle1[LegIndex]+900))*1000/cPwmDiv+cPFConst;
            wTibiaSSCV = ((long)(-TibiaAngle1[LegIndex]+900))*1000/cPwmDiv+cPFConst;
#ifdef c4DOF
            wTarsSSCV = ((long)(-TarsAngle1[LegIndex]+900))*1000/cPwmDiv+cPFConst;
#endif
        } else {
            wCoxaSSCV = ((long)(CoxaAngle1[LegIndex] +900))*1000/cPwmDiv+cPFConst;
            wFemurSSCV = ((long)((long)(FemurAngle1[LegIndex]+900))*1000/cPwmDiv+cPFConst);
            wTibiaSSCV = ((long)(TibiaAngle1[LegIndex]+900))*1000/cPwmDiv+cPFConst;
#ifdef c4DOF
            wTarsSSCV = ((long)(TarsAngle1[LegIndex]+900))*1000/cPwmDiv+cPFConst;
#endif
        }

#ifdef cSSC_BINARYMODE
        SSCSerial.write(pgm_read_byte(&cCoxaPin[LegIndex])  + 0x80);
        SSCSerial.write(wCoxaSSCV >> 8);
        SSCSerial.write(wCoxaSSCV & 0xff);
        SSCSerial.write(pgm_read_byte(&cFemurPin[LegIndex]) + 0x80);
        SSCSerial.write(wFemurSSCV >> 8);
        SSCSerial.write(wFemurSSCV & 0xff);
        SSCSerial.write(pgm_read_byte(&cTibiaPin[LegIndex]) + 0x80);
        SSCSerial.write(wTibiaSSCV >> 8);
        SSCSerial.write(wTibiaSSCV & 0xff);
#ifdef c4DOF
        if ((byte)pgm_read_byte(&cTarsLength[LegIndex])) {    // We allow mix of 3 and 4 DOF legs...
            SSCSerial.write(pgm_read_byte(&cTarsPin[LegIndex]) + 0x80);
            SSCSerial.write(wTarsSSCV >> 8);
            SSCSerial.write(wTarsSSCV & 0xff);
        }
#endif
#ifdef DEBUG_GAITS
        if (g_fDebugOutput && g_fHexOn && (TravelLengthX || TravelLengthZ)) {
            DBGSerial << _DEC((short)pgm_read_byte(&cCoxaPin[LegIndex])) << "=(" << _DEC(CoxaAngle1[LegIndex]) << ")" << _DEC(wCoxaSSCV) << " ";
            DBGSerial << _DEC((short)pgm_read_byte(&cFemurPin[LegIndex])) << "=(" << _DEC(FemurAngle1[LegIndex]) << ")" << _DEC(wFemurSSCV) << " ";
            DBGSerial << _DEC((short)pgm_read_byte(&cTibiaPin[LegIndex])) << "=(" << _DEC(TibiaAngle1[LegIndex]) << ")" << _DEC(wTibiaSSCV) << " ";
        }    
#endif    
#else
        SSCSerial << "#" << _DEC(pgm_read_byte(&cCoxaPin[LegIndex])) << "P" <<  _DEC(wCoxaSSCV);
        SSCSerial << "#" << _DEC(pgm_read_byte(&cFemurPin[LegIndex])) << "P" <<  _DEC(wFemurSSCV);
        SSCSerial << "#" << _DEC(pgm_read_byte(&cTibiaPin[LegIndex])) << "P" <<  _DEC(wTibiaSSCV);
#ifdef c4DOF
        if ((byte)pgm_read_byte(&cTarsLength[LegIndex])) {
            SSCSerial << "#" << _DEC(pgm_read_byte(&cTarsPin[LegIndex])) << "P" <<  _DEC(wTarsSSCV);
        }
#endif
#endif        
    }  
    AllowControllerInterrupts(true);    // Ok for hserial again...
}


//--------------------------------------------------------------------
//[CommitServoDriver Updates the positions of the servos - This outputs
//         as much of the command as we can without committing it.  This
//         allows us to once the previous update was completed to quickly 
//        get the next command to start
//--------------------------------------------------------------------
void ServoDriver::CommitServoDriver()
{
#ifdef cSSC_BINARYMODE
    byte    abOut[3];
#endif
    
    AllowControllerInterrupts(false);    // If on xbee on hserial tell hserial to not processess...
//    disable(TIMERAINT);

#ifdef cSSC_BINARYMODE
    abOut[0] = 0xA1;
    abOut[1] = ServoMoveTime >> 8;
    abOut[2] = ServoMoveTime & 0xff;
    SSCWrite((byte*)(void*)abOut, 3);        // Let the command begin
#else
      //Send <CR>
    SSCSerial << "T" << _DEC(ServoMoveTime) << "\n\r";
#endif

#ifdef DEBUG_GAITS
    if (g_fDebugOutput && g_fHexOn && (TravelLengthX || TravelLengthZ)) {
	DBGSerial << "T:" << _DEC(ServoMoveTime) << endl; 
    }
#endif	
	


//    enable(TIMERAINT);
    AllowControllerInterrupts(true);    

    PrevServoMoveTime = ServoMoveTime;
}

//--------------------------------------------------------------------
//[FREE SERVOS] Frees all the servos
//--------------------------------------------------------------------
void ServoDriver::FreeServos(void)
{
    AllowControllerInterrupts(false);    // If on xbee on hserial tell hserial to not processess...
//  disable(TIMERAINT);
    for (byte LegIndex = 0; LegIndex < 32; LegIndex++) {
        SSCSerial << "#" << _DEC(LegIndex) << "P0";
    }
    SSCSerial.print("T200\r");
//  enable(TIMERAINT);
    AllowControllerInterrupts(true);    
}


//==============================================================================
// SSC Forwarder - used to allow things like Lynxterm to talk to the SSC-32 
// through the Arduino...  Will see if it is fast enough...
//==============================================================================
void  ServoDriver::SSCForwarder(void) 
{
    int sChar;
    int sPrevChar; 
    DBGSerial.println("SSC Forwarder mode - Enter $<cr> to exit");
    
    for (;;) {
        if ((sChar = DBGSerial.read()) != -1) {
            SSCSerial.write(sChar & 0xff);
            if (((sChar == '\n') || (sChar == '\r')) && (sPrevChar == '$'))
                break;    // exit out of the loop
            sPrevChar = sChar;
        }
        

        if ((sChar = SSCSerial.read()) != -1) {
            DBGSerial.write(sChar & 0xff);
        }
    }
    DBGSerial.println("Exited SSC Forwarder mode");
}

void SSCWrite(byte *pb, int ich)
{
    while (ich-- > 0) 
        SSCSerial.write(*pb++);
    
}


//==============================================================================
// Quick and dirty helper function to read so many bytes in from the SSC with a timeout and an end of character marker...
//==============================================================================
int SSCRead (byte* pb, int cb, word wTimeout, word wEOL)
{
    int ich;
    byte* pbIn = pb;
    unsigned long ulTimeLastChar = micros();
#ifdef DEBUG_IOPINS
    digitalWrite(DEBUG_PIN2, HIGH);
#endif
    while (cb) {
        while (!SSCSerial.available()) {
            // check for timeout
            if ((word)(micros()-ulTimeLastChar) > wTimeout) {
#ifdef DEBUG_IOPINS
                digitalWrite(DEBUG_PIN2, LOW);
#endif
                return (int)(pb-pbIn);
            }    
        }
        ich = SSCSerial.read();
        *pb++ = (byte)ich;
        cb--;

        if ((word)ich == wEOL)
            break;    // we matched so get out of here.
        ulTimeLastChar = micros();    // update to say we received something
    }

#ifdef DEBUG_IOPINS
    digitalWrite(DEBUG_PIN2, LOW);
#endif
    return (int)(pb-pbIn);
}

//==============================================================================
//	FindServoOffsets - Find the zero points for each of our servos... 
// 		Will use the new servo function to set the actual pwm rate and see
//		how well that works...
//==============================================================================
void ServoDriver::FindServoOffsets()
{
}

#ifdef LATER
//==============================================================================
//	FindServoZeroPoints - Find the zero points for each of our servos... 
// 		Will use the new servo function to set the actual pwm rate and see
//		how well that works...
//==============================================================================
void FindServoZeroPoints()
{
    // not clean but...
    int data;
    short sSN = 0; 	
    byte bSSCPinNum;		// which servo number
    boolean fNew = true;	// is this a new servo to work with?
    boolean fExit = false;	// when to exit
    byte szTemp[80];
    byte cbRead;
    int ich;
    short sOffset;
    
#define CNTSERVOS 18
    signed char asOffsets[CNTSERVOS];        // we have 18 servos to find/set offsets for...
    signed char asOffsetsRead[CNTSERVOS];    // array for our read in servos...
    
    for (sSN=0; sSN < CNTSERVOS; sSN++) {
         asOffsets[sSN] = 0;           

        if ((sSN%3) == 0) 
            bSSCPinNum = pgm_read_byte(&cCoxaPin[sSN / 3]);
        else if ((sSN%3) == 1)
            bSSCPinNum = pgm_read_byte(&cFemurPin[sSN / 3]);
        else
            bSSCPinNum = pgm_read_byte(&cTibiaPin[sSN / 3]);

        SSCPrintf("R%d\n", 32+bSSCPinNum);    // Note: Servo offsets start at register 32...
        
        // now read in the current value...  Maybe should use atoi...
        cbRead = SSCRead((byte*)szTemp, sizeof(szTemp), 10000, 13);
        if (cbRead > 0)
            asOffsetsRead[sSN] = atoi((const char *)szTemp);
        ich = sprintf((char*)szTemp, "Servo: %d=%d\n", bSSCPinNum, asOffsetsRead[sSN] + asOffsets[sSN]);
        DBGSerial.write(szTemp, ich);
    }

    // Tell all of the zervos to go to 1500
    for (ich=0; ich < 6; ich++)    // Make sure all of our servos initialize to 0 offset from saved.
        SSCPrintf("#%dP1500#%dP1500#%dP1500\n\r", pgm_read_byte(&cCoxaPin[ich]), pgm_read_byte(&cFemurPin[ich]), pgm_read_byte(&cTibiaPin[ich]));


    // OK lets move all of the servos to their zero point.
    DBGSerial.println("Find Servo Zeros.\nEnt-Exit, +- changes, *-change servo");
    sSN = 0;
    bSSCPinNum = 0;    // take care of warning...

    while(!fExit) {
        if (fNew) {
            // We need to get the current setting for this servo
            // First lets map the ssN to which SSC Pin number.
            // BUGBUG:: Probably cleaner ways to do this...
            if ((sSN%3) == 0) 
                bSSCPinNum = pgm_read_byte(&cCoxaPin[sSN / 3]);
            else if ((sSN%3) == 1)
                bSSCPinNum = pgm_read_byte(&cFemurPin[sSN / 3]);
            else
                bSSCPinNum = pgm_read_byte(&cTibiaPin[sSN / 3]);

            ich = sprintf((char*)szTemp, "Servo: %d=%d\n", bSSCPinNum, asOffsetsRead[sSN] + asOffsets[sSN]);
            DBGSerial.write(szTemp, ich);

	    // Now lets wiggle the servo
            SSCPrintf("#%dP1300T500\n\r", bSSCPinNum);
            delay (500);
            SSCPrintf("#%dP1700T500\n\r", bSSCPinNum);
            delay (500);
            SSCPrintf("#%dP%dT500\n\r", bSSCPinNum, asOffsets[sSN]);
            delay (500);
            fNew = false;
        }
	
	//get user entered data
	data = DBGSerial.read();
	//if data received
	if (data !=-1) 	{
            if (((data >= 10) && (data <=15)) || (data == '$'))
		fExit = true;	// not sure how the keypad will map so give NL, CR, LF... all implies exit

	    else if ((data == '+') || (data == '-')) {
	        if (data == '+') {
		     asOffsets[sSN] += 5;		// increment by 5us max 100
                    if ((asOffsets[sSN] + asOffsetsRead[sSN]) > 100)
                        asOffsets[sSN] = 100 - asOffsetsRead[sSN];
                    
                }
		else {
		     asOffsets[sSN] -= 5;		// increment by 5us
                   if ((asOffsets[sSN] + asOffsetsRead[sSN]) < -100)
                        asOffsets[sSN] = -100 - asOffsetsRead[sSN];
                }

                SSCPrintf("#%dP%dT500\n\r", bSSCPinNum, asOffsets[sSN]);
	    } else if ((data >= '0') && (data <= '6')) {
		// direct enter of which servo to change
		fNew = true;
		sSN = (data - '0') * 3;    // choose which leg to do...
	    } else if (data == '*') {
	        // direct enter of which servo to change
		fNew = true;
		sSN++;
		if (sSN == CNTSERVOS) 
		    sSN = 0;	
	    }
	}
    }
    DBGSerial.print("Find Servo exit ");
    DBGSerial.print("\nSave Changes? Y/N: ");

    //get user entered data
    while ((data = DBGSerial.read()) == -1)
	; 

    if ((data == 'Y') || (data == 'y')) {
        // now loop through and tell the SSC-32 to update the servo offsets... 
        for (sSN=0; sSN < CNTSERVOS; sSN++) {
            if ((sSN%3) == 0) 
                bSSCPinNum = pgm_read_byte(&cCoxaPin[sSN / 3]);
            else if ((sSN%3) == 1)
                bSSCPinNum = pgm_read_byte(&cFemurPin[sSN / 3]);
            else
                bSSCPinNum = pgm_read_byte(&cTibiaPin[sSN / 3]);

            SSCPrintf("R%d=%d\n\r", 32+bSSCPinNum, asOffsetsRead[sSN] + asOffsets[sSN]);
	}
        // Then I need to have the SSC-32 reboot in order to use the new values.
        delay(10);    // give it some time to write stuff out.
        SSCPrintf("GOBOOT\n\r");    // tell the SSC-32 to go into boot mode
        delay(5);        // Give it a little time
        SSCPrintf("g0000\n\r");    // tell it that we are done in the boot section so go run the normall SSC stuff...
        delay(500);                // Give it some time to boot up...
    }
    g_ServoDriver.FreeServos();
}
#endif
#endif
